/*
	Copyright 2014, Marten de Vries

	Licensed under the Apache License, Version 2.0 (the "License");
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.
*/

/*global emit, describe, before, PouchDB, after, it */

"use strict";

var chai = require('chai');
chai.should();

//python-pouchdb-js
require("../");

var CONFIG;
try {
  CONFIG = require("../../../pouchdb/tests/testconfig.json");
} catch (e) {
  CONFIG = {}
}
CONFIG.storage_dir = CONFIG.storage_dir || "dbs";
CONFIG.base_url = CONFIG.base_url || "http://localhost:5984";

var AUTH = null;
if (CONFIG.username) {
  AUTH = {
    username: CONFIG.username,
    password: CONFIG.password
  };
}

var db;
var DB_NAME = "node-python-pouchdb-tests";

function showTest(doc, req) {
  return "Hello World!";
}

function listTest(head, req) {
  throw new Error("Hello World!");
}

function viewTest(doc) {
  emit(doc._id);
}

function validateDocUpdateTest(newDoc, oldDoc, userCtx, secObj) {
  if (newDoc._id === "myid") {
    throw {"forbidden": "Mine!!!"};
  }
}

function updateTest(doc, req) {
  return [{_id: req.uuid, hello: "world!"}, req.uuid];
}

before(function () {
  db = new PouchDB(CONFIG.storage_dir + "/" + DB_NAME);
  return db.put({
    _id: "_design/test",
    shows: {
      test: showTest.toString()
    },
    lists: {
      test: listTest.toString()
    },
    views: {
      ids: {
        map: viewTest.toString()
      }
    },
    updates: {
      test: updateTest.toString()
    },
    rewrites: [
      {
        from: "/test/all",
        to: "_list/test/ids"
      }
    ],
    validate_doc_update: validateDocUpdateTest.toString()
  });
});
after(function () {
  return db.destroy();
});

function checkListError(err) {
  err.name.should.equal("Error");
  err.message.should.contain("Hello World!");
  err.status.should.equal(500);
}


describe("plugin tests", function () {
  it("should give the show result", function () {
    return db.show("test/test").then(function (resp) {
      resp.body.should.equal("Hello World!")
      resp.code.should.equal(200);
    });
  });
  it("should get the show result from couchdb", function () {
    var http = new PouchDB(CONFIG.base_url + "/" + DB_NAME, {auth: AUTH});
    return http.replicate.from(db).then(function () {
      return db.show("test/test");
    }).then(function (resp) {
      resp.body.should.equal("Hello World!")
      resp.code.should.equal(200);
    }).then(function () {
      http.destroy();
    });
  });
  it("should give the list error", function () {
    return db.list("test/test/ids").catch(function (err) {
      checkListError(err);
    });
  });
  it("should reject the document", function () {
    return db.validatingPut({"_id": "myid"}).catch(function (err) {
      err.name.should.equal("forbidden");
      err.message.should.equal("Mine!!!");
      err.status.should.equal(403);
    });
  });
  it("should accept the document", function () {
    return db.validatingPost({}).then(function (resp) {
      resp.ok.should.be.ok;
      resp.id.should.be.ok;
      resp.rev.should.be.ok;
    });
  });
  it("should save a document via the update handler", function () {
    return db.update("test/test").then(function (resp) {
      return db.get(resp.body);
    }).then(function (resp) {
      resp.hello.should.equal("world!");
    });
  });
  it("should rewrite the url", function () {
    return db.rewriteResultRequestObject("test/test/all", {query: {"k": "v"}}).then(function (req) {
      req.raw_path.should.equal("/" + encodeURIComponent(CONFIG.storage_dir + "/" + DB_NAME) + "/_design/test/_list/test/ids?k=v");
    });
  });
  it("should return the list function result", function () {
    return db.rewrite("test/test/all").catch(function (err) {
      checkListError(err);
    });
  });
});

